import typing
import warnings
from dataclasses import dataclass
from typing import Dict, List, NamedTuple, Optional, Set, Union

import luigi

from gokart.task import TaskOnKart


@dataclass
class TaskInfo:
    name: str
    unique_id: str
    output_paths: List[TaskOnKart]
    params: dict
    processing_time: str
    is_complete: str
    task_log: dict
    requires: Union['RequiredTask', List['RequiredTask'], Dict[str, 'RequiredTask']]
    children_task_infos: List['TaskInfo']

    def get_task_id(self):
        return f'{self.name}_{self.unique_id}'

    def get_task_title(self):
        return f'({self.is_complete}) {self.name}[{self.unique_id}]'

    def get_task_detail(self):
        return f'(parameter={self.params}, output={self.output_paths}, time={self.processing_time}, task_log={self.task_log})'

    def task_info_dict(self):
        return dict(
            name=self.name,
            unique_id=self.unique_id,
            output_paths=self.output_paths,
            params=self.params,
            processing_time=self.processing_time,
            is_complete=self.is_complete,
            task_log=self.task_log,
            requires=self.requires,
        )


class RequiredTask(NamedTuple):
    name: str
    unique_id: str


def _make_requires_info(requires):
    if isinstance(requires, TaskOnKart):
        return RequiredTask(name=requires.__class__.__name__, unique_id=requires.make_unique_id())
    elif isinstance(requires, dict):
        return {key: _make_requires_info(requires=item) for key, item in requires.items()}
    elif isinstance(requires, typing.Iterable):
        return [_make_requires_info(requires=item) for item in requires]

    raise TypeError(f'`requires` has unexpected type {type(requires)}. Must be `TaskOnKart`, `Iterarble[TaskOnKart]`, or `Dict[str, TaskOnKart]`')


def make_task_info_tree(task: TaskOnKart, ignore_task_names: Optional[List[str]] = None, cache: Optional[Dict[str, TaskInfo]] = None) -> TaskInfo:
    with warnings.catch_warnings():
        warnings.filterwarnings(action='ignore', message='Task .* without outputs has no custom complete() method')
        is_task_complete = task.complete()

    name = task.__class__.__name__
    unique_id = task.make_unique_id()
    output_paths = [t.path() for t in luigi.task.flatten(task.output())]

    cache = {} if cache is None else cache
    cache_id = f'{name}_{unique_id}_{is_task_complete}'
    if cache_id in cache:
        return cache[cache_id]

    params = task.get_info(only_significant=True)
    processing_time = task.get_processing_time()
    if isinstance(processing_time, float):
        processing_time = str(processing_time) + 's'
    is_complete = 'COMPLETE' if is_task_complete else 'PENDING'
    task_log = dict(task.get_task_log())
    requires = _make_requires_info(task.requires())

    children = luigi.task.flatten(task.requires())
    children_task_infos: List[TaskInfo] = []
    for child in children:
        if ignore_task_names is None or child.__class__.__name__ not in ignore_task_names:
            children_task_infos.append(make_task_info_tree(child, ignore_task_names=ignore_task_names, cache=cache))
    task_info = TaskInfo(
        name=name,
        unique_id=unique_id,
        output_paths=output_paths,
        params=params,
        processing_time=processing_time,
        is_complete=is_complete,
        task_log=task_log,
        requires=requires,
        children_task_infos=children_task_infos,
    )
    cache[cache_id] = task_info
    return task_info


def make_tree_info(task_info: TaskInfo, indent: str, last: bool, details: bool, abbr: bool, visited_tasks: Set[str]):
    result = '\n' + indent
    if last:
        result += '└─-'
        indent += '   '
    else:
        result += '|--'
        indent += '|  '
    result += task_info.get_task_title()

    if abbr:
        task_id = task_info.get_task_id()
        if task_id not in visited_tasks:
            visited_tasks.add(task_id)
        else:
            result += f'\n{indent}└─- ...'
            return result

    if details:
        result += task_info.get_task_detail()

    children = task_info.children_task_infos
    for index, child in enumerate(children):
        result += make_tree_info(child, indent, (index + 1) == len(children), details=details, abbr=abbr, visited_tasks=visited_tasks)
    return result


def make_tree_info_table_list(task_info: TaskInfo, visited_tasks: Set[str]):
    task_id = task_info.get_task_id()
    if task_id in visited_tasks:
        return []
    visited_tasks.add(task_id)

    result = [task_info.task_info_dict()]

    children = task_info.children_task_infos
    for child in children:
        result += make_tree_info_table_list(task_info=child, visited_tasks=visited_tasks)
    return result



from abc import abstractmethod
from logging import getLogger
from typing import Any, Dict

import luigi
import numpy as np
import pandas as pd
from luigi.task_register import Register

logger = getLogger(__name__)


class PandasTypeError(Exception):
    pass


class PandasTypeConfig(luigi.Config):
    @classmethod
    @abstractmethod
    def type_dict(cls) -> Dict[str, Any]:
        pass

    @classmethod
    def check(cls, df: pd.DataFrame):
        for column_name, column_type in cls.type_dict().items():
            cls._check_column(df, column_name, column_type)

    @classmethod
    def _check_column(cls, df, column_name, column_type):
        if column_name not in df.columns:
            return

        if not np.all(list(map(lambda x: isinstance(x, column_type), df[column_name]))):
            not_match = next(filter(lambda x: not isinstance(x, column_type), df[column_name]))
            raise PandasTypeError(f'expected type is "{column_type}", but "{type(not_match)}" is passed in column "{column_name}".')


class PandasTypeConfigMap(luigi.Config):
    """To initialize this class only once, this inherits luigi.Config."""

    def __init__(self, *args, **kwargs) -> None:
        super(PandasTypeConfigMap, self).__init__(*args, **kwargs)
        task_names = Register.task_names()
        task_classes = [Register.get_task_cls(task_name) for task_name in task_names]
        self._map = {
            task_class.task_namespace: task_class for task_class in task_classes if issubclass(task_class, PandasTypeConfig) and task_class != PandasTypeConfig
        }

    def check(self, obj, task_namespace: str):
        if isinstance(obj, pd.DataFrame) and task_namespace in self._map:
            self._map[task_namespace].check(obj)



from logging import getLogger
from typing import List, Optional, Set

import luigi

from gokart.task import TaskOnKart
from gokart.tree.task_info import make_task_info_as_tree_str

logger = getLogger(__name__)


def make_tree_info(
    task: TaskOnKart,
    indent: str = '',
    last: bool = True,
    details: bool = False,
    abbr: bool = True,
    visited_tasks: Optional[Set[str]] = None,
    ignore_task_names: Optional[List[str]] = None,
) -> str:
    """
    Return a string representation of the tasks, their statuses/parameters in a dependency tree format

    This function has moved to `gokart.tree.task_info.make_task_info_as_tree_str`.
    This code is remained for backward compatibility.

    Parameters
    ----------
    - task: TaskOnKart
        Root task.
    - details: bool
        Whether or not to output details.
    - abbr: bool
        Whether or not to simplify tasks information that has already appeared.
    - ignore_task_names: Optional[List[str]]
        List of task names to ignore.
    Returns
    -------
    - tree_info : str
        Formatted task dependency tree.
    """
    return make_task_info_as_tree_str(task=task, details=details, abbr=abbr, ignore_task_names=ignore_task_names)


class tree_info(TaskOnKart):
    mode = luigi.Parameter(default='', description='This must be in ["simple", "all"].')  # type: str
    output_path = luigi.Parameter(default='tree.txt', description='Output file path.')  # type: str

    def output(self):
        return self.make_target(self.output_path, use_unique_id=False)



from functools import partial

import luigi


class RunWithLock:
    def __init__(self, func):
        self._func = func

    def __call__(self, instance):
        instance._lock_at_dump = False
        output_list = luigi.task.flatten(instance.output())
        return self._run_with_lock(partial(self._func, self=instance), output_list)

    def __get__(self, instance, owner_class):
        return partial(self.__call__, instance)

    @classmethod
    def _run_with_lock(cls, func, output_list: list):
        if len(output_list) == 0:
            return func()

        output = output_list.pop()
        wrapped_func = output.wrap_with_run_lock(func)
        return cls._run_with_lock(func=wrapped_func, output_list=output_list)



import pandas as pd


def assert_frame_contents_equal(actual: pd.DataFrame, expected: pd.DataFrame, **kwargs):
    """
    Assert that two DataFrames are equal.
    This function is mostly same as pandas.testing.assert_frame_equal, however
    - this fuction ignores the order of index and columns.
    - this function fails when duplicated index or columns are found.

    Parameters
    ----------
    - actual, expected: pd.DataFrame
        DataFrames to be compared.
    - kwargs: Any
        Parameters passed to pandas.testing.assert_frame_equal.
    """
    assert isinstance(actual, pd.DataFrame), 'actual is not a DataFrame'
    assert isinstance(expected, pd.DataFrame), 'expected is not a DataFrame'

    assert actual.index.is_unique, 'actual index is not unique'
    assert expected.index.is_unique, 'expected index is not unique'
    assert actual.columns.is_unique, 'actual columns is not unique'
    assert expected.columns.is_unique, 'expected columns is not unique'

    assert set(actual.columns) == set(expected.columns), 'columns are not equal'
    assert set(actual.index) == set(expected.index), 'indexes are not equal'

    expected_reindexed = expected.reindex(actual.index)[actual.columns]
    pd.testing.assert_frame_equal(actual, expected_reindexed, **kwargs)



from logging import getLogger

import slack_sdk

logger = getLogger(__name__)


class ChannelListNotLoadedError(RuntimeError):
    pass


class ChannelNotFoundError(RuntimeError):
    pass


class FileNotUploadedError(RuntimeError):
    pass


class SlackAPI(object):
    def __init__(self, token, channel: str, to_user: str) -> None:
        self._client = slack_sdk.WebClient(token=token)
        self._channel_id = self._get_channel_id(channel)
        self._to_user = to_user if to_user == '' or to_user.startswith('@') else '@' + to_user

    def _get_channel_id(self, channel_name):
        params = {'exclude_archived': True, 'limit': 100}
        try:
            for channels in self._client.conversations_list(params=params):
                if not channels:
                    raise ChannelListNotLoadedError('Channel list is empty.')
                for channel in channels.get('channels', []):
                    if channel['name'] == channel_name:
                        return channel['id']
            raise ChannelNotFoundError(f'Channel {channel_name} is not found in public channels.')
        except Exception as e:
            logger.warning(f'The job will start without slack notification: {e}')

    def send_snippet(self, comment, title, content):
        try:
            request_body = dict(
                channels=self._channel_id, initial_comment=f'<{self._to_user}> {comment}' if self._to_user else comment, content=content, title=title
            )
            response = self._client.api_call('files.upload', data=request_body)
            if not response['ok']:
                raise FileNotUploadedError(f'Error while uploading file. The error reason is "{response["error"]}".')
        except Exception as e:
            logger.warning(f'Failed to send slack notification: {e}')



import luigi


class SlackConfig(luigi.Config):
    token_name = luigi.Parameter(default='SLACK_TOKEN', description='slack token environment variable.')
    channel = luigi.Parameter(default='', significant=False, description='channel name for notification.')
    to_user = luigi.Parameter(default='', significant=False, description='Optional; user name who is supposed to be mentioned.')
    send_tree_info = luigi.BoolParameter(
        default=False,
        significant=False,
        description='When this option is true, the dependency tree of tasks is included in send message.'
        'It is recommended to set false to this option when notification takes long time.',
    )



import itertools
import os
import pathlib
from logging import getLogger

import luigi

import gokart

logger = getLogger(__name__)


def _get_all_output_file_paths(task: gokart.TaskOnKart):
    output_paths = [t.path() for t in luigi.task.flatten(task.output())]
    children = luigi.task.flatten(task.requires())
    output_paths.extend(itertools.chain.from_iterable([_get_all_output_file_paths(child) for child in children]))
    return output_paths


def delete_local_unnecessary_outputs(task: gokart.TaskOnKart):
    task.make_unique_id()  # this is required to make unique ids.
    all_files = {str(path) for path in pathlib.Path(task.workspace_directory).rglob('*.*')}
    log_files = {str(path) for path in pathlib.Path(os.path.join(task.workspace_directory, 'log')).rglob('*.*')}
    necessary_files = set(_get_all_output_file_paths(task))
    unnecessary_files = all_files - necessary_files - log_files
    if len(unnecessary_files) == 0:
        logger.info('all files are necessary for this task.')
    else:
        logger.info(f'remove following files: {os.linesep} {os.linesep.join(unnecessary_files)}')
    for file in unnecessary_files:
        os.remove(file)
